from abc import ABCMeta, abstractmethod
from types import TracebackType
from typing import Callable, Coroutine, Optional, Type, TypeVar

T_Retval = TypeVar('T_Retval')


class TaskGroup(metaclass=ABCMeta):
    """
    Groups several asynchronous tasks together.

    :ivar cancel_scope: the cancel scope inherited by all child tasks
    :vartype cancel_scope: CancelScope
    """

    cancel_scope: 'CancelScope'

    @abstractmethod
    async def spawn(self, func: Callable[..., Coroutine], *args, name=None) -> None:
        """
        Launch a new task in this task group.

        :param func: a coroutine function
        :param args: positional arguments to call the function with
        :param name: name of the task, for the purposes of introspection and debugging
        """

    @abstractmethod
    async def __aenter__(self) -> 'TaskGroup':
        """Enter the task group context and allow starting new tasks."""

    @abstractmethod
    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> Optional[bool]:
        """Exit the task group context waiting for all tasks to finish."""


class CancelScope(metaclass=ABCMeta):
    @abstractmethod
    async def cancel(self) -> None:
        """Cancel this scope immediately."""

    @property
    @abstractmethod
    def deadline(self) -> float:
        """
        The time (clock value) when this scope is cancelled automatically.

        Will be ``float('inf')``` if no timeout has been set.
        """

    @property
    @abstractmethod
    def cancel_called(self) -> bool:
        """``True`` if :meth:`cancel` has been called."""

    @property
    @abstractmethod
    def shield(self) -> bool:
        """
        ``True`` if this scope is shielded from external cancellation.

        While a scope is shielded, it will not receive cancellations from outside.
        """

    @abstractmethod
    async def __aenter__(self):
        pass

    @abstractmethod
    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> Optional[bool]:
        pass



import codecs
from dataclasses import InitVar, dataclass, field
from typing import Callable, Tuple

from ..abc import (
    AnyByteReceiveStream, AnyByteSendStream, ObjectReceiveStream, ObjectSendStream, ObjectStream)
from ..abc.streams import AnyByteStream


@dataclass
class TextReceiveStream(ObjectReceiveStream[str]):
    """
    Stream wrapper that decodes bytes to strings using the given encoding.

    Receives from the wrapped stream until the given delimiter is found, and then returns the
    string (without the delimiter).

    :param transport_stream: any bytes-based receive stream
    :param encoding: character encoding to use for decoding bytes to strings (defaults to
        ``utf-8``)
    :param errors: handling scheme for decoding errors (defaults to ``strict``; see the
        `codecs module documentation`_ for a comprehensive list of options)

    .. _codecs module documentation: https://docs.python.org/3/library/codecs.html#codec-objects
    """

    transport_stream: AnyByteReceiveStream
    encoding: InitVar[str] = 'utf-8'
    errors: InitVar[str] = 'strict'
    _decoder: codecs.IncrementalDecoder = field(init=False)

    def __post_init__(self, encoding, errors):
        decoder_class = codecs.getincrementaldecoder(encoding)
        self._decoder = decoder_class(errors=errors)

    async def receive(self) -> str:
        while True:
            chunk = await self.transport_stream.receive()
            decoded = self._decoder.decode(chunk)
            if decoded:
                return decoded

    async def aclose(self) -> None:
        await self.transport_stream.aclose()
        self._decoder.reset()

    @property
    def extra_attributes(self):
        return self.transport_stream.extra_attributes


@dataclass
class TextSendStream(ObjectSendStream[str]):
    """
    Sends strings to the wrapped stream as bytes using the given encoding.

    :param AnyByteSendStream transport_stream: any bytes-based send stream
    :param str encoding: character encoding to use for encoding strings to bytes (defaults to
        ``utf-8``)
    :param str errors: handling scheme for encoding errors (defaults to ``strict``; see the
        `codecs module documentation`_ for a comprehensive list of options)

    .. _codecs module documentation: https://docs.python.org/3/library/codecs.html#codec-objects
    """

    transport_stream: AnyByteSendStream
    encoding: InitVar[str] = 'utf-8'
    errors: str = 'strict'
    _encoder: Callable[..., Tuple[bytes, int]] = field(init=False)

    def __post_init__(self, encoding):
        self._encoder = codecs.getencoder(encoding)

    async def send(self, item: str) -> None:
        encoded = self._encoder(item, self.errors)[0]
        await self.transport_stream.send(encoded)

    async def aclose(self) -> None:
        await self.transport_stream.aclose()

    @property
    def extra_attributes(self):
        return self.transport_stream.extra_attributes


@dataclass
class TextStream(ObjectStream[str]):
    """
    A bidirectional stream that decodes bytes to strings on receive and encodes strings to bytes on
    send.

    Extra attributes will be provided from both streams, with the receive stream providing the
    values in case of a conflict.

    :param AnyByteStream transport_stream: any bytes-based stream
    :param str encoding: character encoding to use for encoding/decoding strings to/from bytes
        (defaults to ``utf-8``)
    :param str errors: handling scheme for encoding errors (defaults to ``strict``; see the
        `codecs module documentation`_ for a comprehensive list of options)

    .. _codecs module documentation: https://docs.python.org/3/library/codecs.html#codec-objects
    """

    transport_stream: AnyByteStream
    encoding: InitVar[str] = 'utf-8'
    errors: InitVar[str] = 'strict'
    _receive_stream: TextReceiveStream = field(init=False)
    _send_stream: TextSendStream = field(init=False)

    def __post_init__(self, encoding, errors):
        self._receive_stream = TextReceiveStream(self.transport_stream, encoding=encoding,
                                                 errors=errors)
        self._send_stream = TextSendStream(self.transport_stream, encoding=encoding, errors=errors)

    async def receive(self) -> str:
        return await self._receive_stream.receive()

    async def send(self, item: str) -> None:
        await self._send_stream.send(item)

    async def send_eof(self) -> None:
        await self.transport_stream.send_eof()

    async def aclose(self) -> None:
        await self._send_stream.aclose()
        await self._receive_stream.aclose()

    @property
    def extra_attributes(self):
        return dict(**self.send_stream.extra_attributes, **self.receive_stream.extra_attributes)



import threading
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future
from typing import Any, Callable, Coroutine, TypeVar, overload

T_Retval = TypeVar('T_Retval')


class BlockingPortal(metaclass=ABCMeta):
    """An object tied that lets external threads run code in an asynchronous event loop."""

    __slots__ = '_task_group', '_event_loop_thread_id', '_stop_event', '_cancelled_exc_class'

    def __init__(self):
        from .. import create_event, create_task_group, get_cancelled_exc_class

        self._event_loop_thread_id = threading.get_ident()
        self._stop_event = create_event()
        self._task_group = create_task_group()
        self._cancelled_exc_class = get_cancelled_exc_class()

    async def __aenter__(self) -> 'BlockingPortal':
        await self._task_group.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()
        return await self._task_group.__aexit__(exc_type, exc_val, exc_tb)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.call(self.stop, exc_type is not None)

    async def sleep_until_stopped(self) -> None:
        """Sleep until :meth:`stop` is called."""
        await self._stop_event.wait()

    async def stop(self, cancel_remaining: bool = False) -> None:
        """
        Signal the portal to shut down.

        This marks the portal as no longer accepting new calls and exits from
        :meth:`sleep_until_stopped`.

        :param cancel_remaining: ``True`` to cancel all the remaining tasks, ``False`` to let them
            finish before returning

        """
        self._event_loop_thread_id = None
        await self._stop_event.set()
        if cancel_remaining:
            await self._task_group.cancel_scope.cancel()

    def stop_from_external_thread(self, cancel_remaining: bool = False) -> None:
        """
        Signal the portal to stop and wait for the event loop thread to finish.

        :param cancel_remaining: ``True`` to cancel all the remaining tasks, ``False`` to let them
            finish before returning

        """
        thread = self.call(threading.current_thread)
        self.call(self.stop, cancel_remaining)
        thread.join()

    async def _call_func(self, func: Callable, args: tuple, future: Future) -> None:
        try:
            retval = func(*args)
            if isinstance(retval, Coroutine):
                future.set_result(await retval)
            else:
                future.set_result(retval)
        except self._cancelled_exc_class:
            future.cancel()
        except BaseException as exc:
            future.set_exception(exc)

    @abstractmethod
    def _spawn_task_from_thread(self, func: Callable, args: tuple, future: Future) -> None:
        pass

    @overload
    def call(self, func: Callable[..., Coroutine[Any, Any, T_Retval]], *args) -> T_Retval:
        ...

    @overload
    def call(self, func: Callable[..., T_Retval], *args) -> T_Retval:
        ...

    def call(self, func, *args):
        """
        Call the given function in the event loop thread.

        If the callable returns a coroutine object, it is awaited on.

        :param func: any callable

        :raises RuntimeError: if this method is called from within the event loop thread

        """
        if self._event_loop_thread_id is None:
            raise RuntimeError('This portal is not running')
        if self._event_loop_thread_id == threading.get_ident():
            raise RuntimeError('This method cannot be called from the event loop thread')

        future: Future = Future()
        self._spawn_task_from_thread(func, args, future)
        return future.result()



#  tests for anyio-2.0.2-py39h6ffa863_2 (this is a generated file);
print('===== testing package: anyio-2.0.2-py39h6ffa863_2 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
import sys

import pytest

print("preflighting uvloop...", flush=True)
try:
    import uvloop
    print("... uvloop _appears_ to be working!")
except Exception as err:
    print(f"... uvloop import error, ignoring:\n{err}")

COV_THRESHOLD = {
    "win32": "68",
    "linux": "71",
    "darwin": "71"
}.get(sys.platform)

SKIPS = [
    "connection_refused",
    "happy_eyeballs",
    "ipv6",
]

PYTEST_ARGS = [
    "-vv",
    "--cov", os.environ["PKG_NAME"],
    "--cov-fail-under", COV_THRESHOLD,
    "-k", f"""not ({" or ".join(SKIPS)})""",
    "src"
]

print(">>> pytest", " ".join(PYTEST_ARGS), flush=True)

sys.exit(pytest.main(PYTEST_ARGS))
#  --- run_test.py (end) ---

print('===== anyio-2.0.2-py39h6ffa863_2 OK =====');
print("import: 'anyio'")
import anyio




import pytest

from anyio import open_file

pytestmark = pytest.mark.anyio


@pytest.fixture(scope='module')
def testdata():
    return b''.join(bytes([i] * 1000) for i in range(10))


@pytest.fixture
def testdatafile(tmp_path_factory, testdata):
    file = tmp_path_factory.mktemp('file').joinpath('testdata')
    file.write_bytes(testdata)
    return file


async def test_open_close(testdatafile):
    f = await open_file(testdatafile)
    await f.aclose()


async def test_read(testdatafile, testdata):
    async with await open_file(testdatafile, 'rb') as f:
        data = await f.read()

    assert f.closed
    assert data == testdata


async def test_write(testdatafile, testdata):
    async with await open_file(testdatafile, 'ab') as f:
        await f.write(b'f' * 1000)

    assert testdatafile.stat().st_size == len(testdata) + 1000


async def test_async_iteration(tmp_path):
    lines = ['blah blah\n', 'foo foo\n', 'bar bar']
    testpath = tmp_path.joinpath('testfile')
    testpath.write_text(''.join(lines), 'ascii')
    async with await open_file(str(testpath)) as f:
        lines_i = iter(lines)
        async for line in f:
            assert line == next(lines_i)
